import QtQuick 1.1
import QtWidgetProxies 1.0

FocusScope
{
	id: virtualList
	property variant model
	property variant delegate
	property variant sectionDelegate
    property int contentY: 0
	// a list of height of all the groups
    property variant itemHeights
    property int sectionHeight 
	property int wheelAnimationDuration: 500
	property real wheelAnimationMaxOvershoot: 30
	property int tabOrder: 1

	property variant currentIndex

	onCurrentIndexChanged:
	{
		if (currentIndex && currentIndex.item && !currentIndex.item.activeFocus)
			currentIndex.item.forceActiveFocus();
	}
	onActiveFocusChanged:
	{
		if (activeFocus && !currentIndex && topItemDescriptor)
			moveFocusToNearestIndex(topItemDescriptor, true);
	}
	function indexFromDescription(description)
	{
		var array = (description.itemIndex == -1) ? sectionDescriptors : itemDescriptors;

		for (var i = 0; i < array.length; ++i)
		{
			var index = array[i];
			if (areItemDescriptorsEqual(description, index))
				return index;
		}
		return description;
	}
	function updateScreenForIndex(descriptor)
	{
		var endY = descriptor.startY + ((descriptor.itemIndex == -1) ? sectionHeight : itemHeights[descriptor.groupIndex]);
		if (endY > contentY + virtualList.height)
			contentY = endY - virtualList.height;
		else if (descriptor.startY < contentY)
			contentY = descriptor.startY - ((descriptor.itemIndex == -1) ? 0 : sectionHeight);
	}
	function moveFocusToNextIndex(index, isForward, shiftScreenIsAllowed)
	{
		var nextDescriptor = isForward ? getNextItem(index) : getPreviousItem(index);
		if ((!nextDescriptor || (nextDescriptor.groupIndex == -1 && nextDescriptor.itemIndex == -1 )))
		{
			if (shiftScreenIsAllowed)
				updateScreenForIndex(index);
			return;
		}
		
		if (shiftScreenIsAllowed)
			updateScreenForIndex(nextDescriptor);
		
		var nextIndex = indexFromDescription(nextDescriptor);
		if (nextIndex.item.canHaveFocus)
			currentIndex = nextIndex;
		else
			moveFocusToNextIndex(nextIndex, isForward, shiftScreenIsAllowed);
	}
	function moveFocusToNearestIndex(index, isForward)
	{
		if (index.item && index.item.canHaveFocus)
			currentIndex = index;
		else
			moveFocusToNextIndex(index, isForward, false);
	}
	function onPageDown()
	{
		if (contentY + 2 * height < contentHeight)
			contentY = contentY + height;
		else if (contentY + height < contentHeight)
			contentY = contentHeight - height;
		var index = itemDescriptors[itemDescriptors.length - 1];
		if (index.startY + itemHeights[index.groupIndex] > contentY + height && itemDescriptors.length > 1)
			index = itemDescriptors[itemDescriptors.length - 2];
		moveFocusToNearestIndex(index, false);
	}
	function onPageUp()
	{
		contentY = (contentY - height > 0) ? (contentY - height) : 0;
		moveFocusToNearestIndex(topItemDescriptor, true);
	}
	
	function onEnd()
	{
		contentY = (contentHeight > height) ? (contentHeight - height) : 0
		if (itemDescriptors.length)
			moveFocusToNearestIndex(itemDescriptors[itemDescriptors.length - 1], false);
	}
	function onHome()
	{
		contentY = 0;
		moveFocusToNearestIndex(topItemDescriptor, true);
	}
	
	Keys.onPressed:
	{
		if (event.modifiers & (Qt.ShiftModifier | Qt.ControlModifier | Qt.AltModifier | Qt.MetaModifier))
		{
			event.accepted = false;
			return;
		}
			
		switch(event.key)
		{
			case Qt.Key_Down: moveFocusToNextIndex(currentIndex, true, true); break;
			case Qt.Key_Up: moveFocusToNextIndex(currentIndex, false, true); break;
			case Qt.Key_PageDown: onPageDown(); break;
			case Qt.Key_PageUp: onPageUp(); break;
			case Qt.Key_Home: onHome(); break;
			case Qt.Key_End: onEnd(); break;
			default: event.accepted = false;
		}
	}

    property int contentHeight
	property variant visibleArea: QtObject {
		property real yPosition: contentY / contentHeight
	}

	// private properties

	property list<Item> itemPool
	property list<Item> sectionPool
	property variant topItemDescriptor
	property variant itemDescriptors: []
	property variant sectionDescriptors: []

	Connections
	{
		target: model
		ignoreUnknownSignals: false

        onGroupRefresh: {
            updateContentHeight();
            itemDescriptors = [];
            var newContentY = contentY + height > contentHeight 
                ? Math.max(contentHeight - height, 0)
                : contentY
            if (newContentY === contentY)
                realign();
            else
                contentY = newContentY;
        }
		
		onItemAdded: {
			var addedItemDescriptor = {groupIndex: groupId, itemIndex: row};
			updateContentHeight();
			if (isItemDescriptorLess(addedItemDescriptor, topItemDescriptor))
			{
				contentY = getItemPosition(topItemDescriptor) + (contentY - topItemDescriptor.startY);
			}
			else if (getItemPosition(addedItemDescriptor) < contentY + height)
			{
				itemDescriptors = [];
				realign();
			}
		}
		
		onItemRemoved: {
			var newContentY;
			updateContentHeight();
			if (isItemDescriptorLess({groupIndex: groupId, itemIndex: row}, topItemDescriptor))
			{
				contentY = getItemPosition(topItemDescriptor) + (contentY - topItemDescriptor.startY);
			}
			else if (contentY + height > contentHeight)
			{
				itemDescriptors = [];
				newContentY = Math.max(contentHeight - height, 0);
				if (contentY == newContentY)
				    realign();
				else
					contentY = newContentY; 				
			}
			else
			{
				itemDescriptors = [];
				realign();
			}
		}
		
		
		onItemUpdated: {
			var updatedItemY = getItemPosition({groupIndex: groupId, itemIndex: row});
			if (updatedItemY < contentY + height && updatedItemY + itemHeights[groupId] >= contentY)
			{
				itemDescriptors = [];
				realign();
			}
		}	
	}
    
	onDelegateChanged: {
		if (status == Component.Ready)
		{
			recreateItemPool();
			realign();
		}
	}
    onContentYChanged: {
		if (status == Component.Ready)
			realign();
	}
    onModelChanged: {
		if (status == Component.Ready)
		{
			updateContentHeight();
			recreateSectionPool();
			realign();
		}
    }
    onItemHeightsChanged: {
		if (status == Component.Ready)
		{
			updateContentHeight();
			recreateItemPool();
			realign();
		}
    }
	onHeightChanged: {
		if (status == Component.Ready)
		{
			recreateItemPool();
			realign();
		}
	}
	onWidthChanged: {		
		if (status == Component.Ready)
		{
			recreateItemPool();
			realign();
		}
	}	
    onSectionHeightChanged: {
		if (status == Component.Ready)
		{
			updateContentHeight();
			recreateSectionPool();
			realign();
		}
    }
	Component.onCompleted: {
		recreateSectionPool();
		recreateItemPool();
		realign();
	}
    
	PropertyAnimation {
         id: mouseWheelAnimation
         target: virtualList
         properties: "contentY"
         duration: wheelAnimationDuration
    }

	MouseWheelArea
	{
		anchors.fill: parent
		
		onVerticalWheel: {
			var newContentY = Math.floor(
				(mouseWheelAnimation.running ? mouseWheelAnimation.to : contentY) - delta);
			var overshoot = 0;
			if (newContentY < 0 ) 
			{
				overshoot = -newContentY;
				newContentY = 0;				
			}
			else if (newContentY >= contentHeight - height) 
			{
				overshoot = newContentY - contentHeight + height;
				newContentY = Math.max(0, contentHeight - height);				
			}
			if (overshoot > wheelAnimationMaxOvershoot)
				overshoot = wheelAnimationMaxOvershoot;
			mouseWheelAnimation.stop();
			mouseWheelAnimation.to = newContentY;
			mouseWheelAnimation.easing.type = overshoot > 0 ? Easing.OutBack : Easing.InOutQuad;
			mouseWheelAnimation.easing.overshoot = overshoot / 12;
			mouseWheelAnimation.start();
		}
	}

    function destroyList(list)
    {
        var i;
	    for (i = 0; i < list.length; ++i)
            list[i].destroy();
    }

    function updateContentHeight()
    {
        var i, c = 0;
        if (model != undefined) 
        {
            for (i = 0; i < model.groups.length; ++i)
            {
                if (model.groups[i].Count() > 0)
                {
                    c += sectionHeight;
                    if (itemHeights)
                    	c += model.groups[i].Count() * itemHeights[i];
                }
            }
        }
        contentHeight = c;
    }
	
	function isItemDescriptorLess(left, right)
	{
		if (left.groupIndex < right.groupIndex)
			return true;
		if (left.groupIndex == right.groupIndex)
			return left.itemIndex < right.itemIndex;
		return false;
	}
	
	function areItemDescriptorsEqual(left, right)
	{
		return left.groupIndex == right.groupIndex &&
			left.itemIndex == right.itemIndex;
	}


    function getItemAt(y)
    {
        var i, c = 0, sectionStart, itemsStart, itemIndex;
        if (model != undefined) 
        {
            for (i = 0; i < model.groups.length; ++i)
            {
                if (model.groups[i].Count() > 0)
                {
                    sectionStart = c
                    c += sectionHeight;
                    if (c > y )
                        return { groupIndex: i, itemIndex: -1, startY: sectionStart };
                    itemsStart = c
                    c += model.groups[i].Count() * itemHeights[i];
                    if (c > y)
                    {
                        itemIndex = Math.floor((y - itemsStart) / itemHeights[i])
                        return { groupIndex: i, itemIndex: itemIndex, startY: itemsStart + itemIndex * itemHeights[i] };
                    }
                }
            }
        }
        return { groupIndex: -1, itemIndex: -1, startY: -1 }
    }
	
	function getItemPosition(itemDescriptor)
	{
        var i, c = 0;
        if (model != undefined) 
        {
            for (i = 0; i < itemDescriptor.groupIndex; ++i)
            {
                if (model.groups[i].Count() > 0)
                {
                    c += sectionHeight;
                    c += model.groups[i].Count() * itemHeights[i];
                }
            }
			if (itemDescriptor.itemIndex == -1)
				return c;
			c += sectionHeight;
			return c + itemDescriptor.itemIndex * itemHeights[i];
        }
        return 0;	
	}

    function getNextItem(item)
    {
        var i, c, sectionStart, itemsStart, itemIndex;
        c = 0
        if (model != undefined && item.groupIndex != -1) 
        {
            if (item.itemIndex == -1)
                return { groupIndex: item.groupIndex, itemIndex: 0, startY: item.startY + sectionHeight };
            if (item.itemIndex + 1 < model.groups[item.groupIndex].Count())
                return { groupIndex: item.groupIndex, itemIndex: item.itemIndex + 1, startY: item.startY + itemHeights[item.groupIndex] };
            for (i = item.groupIndex + 1; i < model.groups.length; ++i)
                if (model.groups[i].Count() > 0)
                    return { groupIndex: i, itemIndex: -1, startY: item.startY + itemHeights[item.groupIndex] };
        }
        return { groupIndex: -1, itemIndex: -1, startY: -1 }
    }

    function getPreviousItem(item)
    {
        var i, c, sectionStart, itemsStart, itemIndex;
        c = 0
        if (model != undefined && item.groupIndex != -1) 
        {
            if (item.itemIndex > 0)
                return { groupIndex: item.groupIndex, itemIndex: item.itemIndex - 1, startY: item.startY - itemHeights[item.groupIndex] };
            if (item.itemIndex == 0)
                return { groupIndex: item.groupIndex, itemIndex: item.itemIndex - 1, startY: item.startY - sectionHeight };
            for (i = item.groupIndex - 1; i >= 0; --i)
                if (model.groups[i].Count() > 0)
                    return { groupIndex: i, itemIndex: model.groups[i].Count()-1, startY: item.startY - itemHeights[i] };
        }
        return { groupIndex: -1, itemIndex: -1, startY: -1 }
    }
	

	function recreateSectionPool()
	{
		var newSections = [], i, section;
		if (sectionDelegate == undefined || model == undefined)
			return;
		if (sectionPool.length >= model.groups.length)
			return;		
        destroyList(sectionPool);				
		for (i = 0; i < model.groups.length; ++i)
		{
			section = sectionDelegate.createObject(virtualList);
			section.visible = false
			section.sectionModel = model.groups[i]
			newSections.push(section)
		}
        sectionPool = newSections;
	}
	
	function minInArray(array)
	{
		var i, value;
		for (i = 0; i < array.length; ++i)
			if (value == undefined || array[i] < value)
				value = array[i];
		return value;
	}
	
	function recreateItemPool()
	{
		if (delegate == undefined || height == 0 || !itemHeights )
			return;
		var maxVisibleItems = Math.ceil(height / minInArray(itemHeights)) + 1;
		var newItems = [];
		var i, item;
		if (maxVisibleItems <= itemPool.length)
			return;
        destroyList(itemPool);		
		for (i = 0; i < maxVisibleItems; ++i)
		{
			item = delegate.createObject(virtualList);
			item.visible = false
			newItems.push(item)
		}
        itemPool = newItems;
	}
	
	function realign()
	{
		if (delegate == undefined || sectionDelegate == undefined || height == 0 || !itemHeights || width == 0)
			return;
		var visibleContentEnd = contentY + height;
		var itemDescriptor;
		var sectionsUsed = 0, itemsUsed = 0;
		var visualItem, groupModel, i;
		var newItemDescriptors = [], newSectionDescriptors = [], topSectionDescriptor;
        if (model != undefined)
        {
			topItemDescriptor = getItemAt(contentY)
			for (itemDescriptor = topItemDescriptor;
				itemDescriptor.groupIndex != -1 && itemDescriptor.startY < visibleContentEnd;
				itemDescriptor = getNextItem(itemDescriptor))
            {
				groupModel = model.groups[itemDescriptor.groupIndex]; 
				if (itemDescriptor.startY - contentY <= 0)
				{
					visualItem = sectionPool[sectionsUsed];
					topSectionDescriptor = { groupIndex: itemDescriptor.groupIndex, itemIndex: -1, startY: contentY };
					if (itemDescriptor.itemIndex == groupModel.Count() - 1 && 
						itemDescriptor.startY - contentY + itemHeights[itemDescriptor.groupIndex] < sectionHeight)
					{
						topSectionDescriptor.startY = itemDescriptor.startY  + itemHeights[itemDescriptor.groupIndex] - sectionHeight;
					}
					updateVisualItem(visualItem,topSectionDescriptor);
					newSectionDescriptors[sectionsUsed] = topSectionDescriptor;
					++sectionsUsed;
					if (visualItem)
					{
						bindItem(visualItem, topSectionDescriptor);
						positionItem(visualItem, topSectionDescriptor);
						visualItem.z = 1;
					}
					if (itemDescriptor.itemIndex == -1)
					{
						continue;
					}
				}
				if (itemDescriptor.itemIndex == -1)
				{
					visualItem = sectionPool[sectionsUsed];
					bindItem(visualItem, itemDescriptor);
					updateVisualItem(visualItem, itemDescriptor);
					positionItem(visualItem, itemDescriptor);
					newSectionDescriptors[sectionsUsed] = itemDescriptor;
					++sectionsUsed;
				}
				else				
				{
					visualItem = itemPool[itemsUsed];
					if (itemsUsed >= itemDescriptors.length ||
						!areItemDescriptorsEqual(itemDescriptor, itemDescriptors[itemsUsed]))
					{
						bindItem(visualItem, itemDescriptor);
					}
					updateVisualItem(visualItem, itemDescriptor);
					positionItem(visualItem, itemDescriptor);
					newItemDescriptors[itemsUsed] = itemDescriptor;
					++itemsUsed;
				}
            }
        }
		for (i = sectionsUsed; i < sectionPool.length; ++i)
			sectionPool[i].visible = false;
		for (i = itemsUsed; i < itemPool.length; ++i)
			itemPool[i].visible = false;
		sectionDescriptors = newSectionDescriptors;
		itemDescriptors = newItemDescriptors;		
	}
	function updateVisualItem(visualItem, itemDescriptor)
	{
		if (visualItem)
		{
			itemDescriptor.item = visualItem;
			visualItem.index = itemDescriptor;
		}
	}
    function bindItem(visualItem, itemDescriptor)
    {		
        var groupModel;
        var roles;
		var i;
        if (itemDescriptor.groupIndex != -1)
        {
            groupModel = model.groups[itemDescriptor.groupIndex]; 
            if (itemDescriptor.itemIndex == -1)
            {
            	if (visualItem)
                	visualItem.sectionModel = groupModel
            }
            else
            {
				roles = groupModel.roles;
				var paramMap = {};
				paramMap.groupModel = groupModel;
				for (i = 0; i < roles.length; ++i)
				{
					paramMap[roles[i]] = groupModel.GetData(itemDescriptor.itemIndex, roles[i]);
				}
				visualItem.itemModel = paramMap;
            }
        }
    }
	
	function positionItem(visualItem, itemDescriptor)
    {
    	if (!visualItem || !itemDescriptor)
    		return;
        if (itemDescriptor.groupIndex != -1)
        {
            if (itemDescriptor.itemIndex == -1)
            {
                visualItem.height = sectionHeight;
            }
            else
            {
                visualItem.height = itemHeights[itemDescriptor.groupIndex];
            }
            visualItem.x = 0
            visualItem.width = width
            visualItem.y = itemDescriptor.startY - contentY;
			visualItem.z = 0;
			visualItem.visible = true;
        }
    }


}


